/*
 * Copyright 2020 ETSI
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holder nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.etsi.saref.servers.search;

import java.io.IOException;
import java.util.ArrayList;
import java.util.function.Function;

import javax.ws.rs.BadRequestException;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.InternalServerErrorException;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.MediaType;

import org.apache.jena.query.Dataset;
import org.apache.jena.query.QueryExecutionFactory;
import org.apache.jena.query.QuerySolution;
import org.apache.jena.query.ReadWrite;
import org.apache.jena.rdf.model.RDFNode;
import org.apache.jena.sparql.expr.nodevalue.NodeValueString;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Omar Qawasmeh, maxime.lefrancois
 *
 */
@Path("/search")
public class Search {

	private static final Logger LOG = LoggerFactory.getLogger(Search.class);
	private static final int NB_RESULTS = 20;

	private static final String QUERY_COUNT = "PREFIX ex: <http://example.org/>\n"
			+ "PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>\n" + "PREFIX text: <http://jena.apache.org/text#>\n"
			+ "SELECT (COUNT(*) AS ?count)\n" + "WHERE {\n" + "   ( ?s ?score ?literal ?g ) text:query %s .\n"
			+ "  OPTIONAL { ?s rdfs:label ?label }\n" + "  OPTIONAL { ?s rdfs:comment ?comment }\n" + "}";

	private static final String QUERY = "PREFIX ex: <http://example.org/>\n"
			+ "PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>\n" + "PREFIX text: <http://jena.apache.org/text#>\n"
			+ "SELECT DISTINCT ?s ?label ?comment ?score \n" + "WHERE {\n"
			+ "   ( ?s ?score ?literal ) text:query %s .\n" + "  OPTIONAL { ?s rdfs:label ?label }\n"
			+ "  OPTIONAL { ?s rdfs:comment ?comment }\n" + "}\n" + "ORDER BY DESC ( ?score )\n" + "OFFSET %s\n"
			+ "LIMIT %s";

	

	@GET
	@Produces(MediaType.APPLICATION_JSON)
	public SearchResult doSearch(@QueryParam("q") String keyword,
			@DefaultValue(value = "1") @QueryParam("page") int page)
			throws WebApplicationException, IOException, InterruptedException {
		
		Dataset dataset = DatasetManager.getDataset();
		if(dataset == null) {
			throw new InternalServerErrorException("Could not get a handle on the dataset");
		}

		if (keyword == null) {
			throw new BadRequestException("Expecting query parameter q to be set");
		}

		// build Node to avoid any code injection
		String kw = new NodeValueString(keyword).asQuotedString();

		SearchResult searchResult = new SearchResult();
		searchResult.search = keyword;

		try {
			final String queryCount = String.format(QUERY_COUNT, kw);
			LOG.trace(queryCount);

			dataset.begin(ReadWrite.READ);
			QueryExecutionFactory.create(queryCount, dataset).execSelect().forEachRemaining(action -> {
				searchResult.count = action.get("count").asLiteral().getInt();
			});

			searchResult.offset = (page - 1) * NB_RESULTS;
			searchResult.results = new ArrayList<>();

			final String queryResults = String.format(QUERY, kw, searchResult.offset, searchResult.offset + NB_RESULTS);
			LOG.trace(queryResults);
			QueryExecutionFactory.create(queryResults, dataset).execSelect().forEachRemaining(action -> {
				SearchResult.SearchResultSummary result = new SearchResult.SearchResultSummary();
				result.uri = apply(action, "s", node -> node.asResource().getURI());
				result.score = apply(action, "score", node -> node.asLiteral().getFloat());
				result.label = apply(action, "label", node -> node.asLiteral().getLexicalForm());
				result.comment = apply(action, "comment", node -> node.asLiteral().getLexicalForm());
				searchResult.results.add(result);
			});
			dataset.end();
			return searchResult;
		} catch (Exception ex) {
			LOG.warn("exception while evaluating search", ex);
			throw new InternalServerErrorException(ex);
		}

	}

	private <R> R apply(QuerySolution sol, String variable, Function<RDFNode, R> function) {
		try {
			return function.apply(sol.get(variable));
		} catch (Exception ex) {
			return null;
		}
	}

}
